__all__ = 'ChangeListCustomTemplateNameAdminMixin',


class ChangeListCustomTemplateNameAdminMixin:
    """Simple mixin to made auto-generatable template paths out of template name.

    It mimics default django admin template resolving behavior, but it changes
    default `change_list` template name to thw one from
    `self.change_list_template_name` attribute.
    """
    change_list_template_name = 'change_list'

    @property
    def change_list_template(self):
        name = self.change_list_template_name
        opts = self.model._meta
        app_label = opts.app_label

        return [
            f'admin/{app_label}/{opts.model_name}/{name}.html',
            f'admin/{app_label}/{name}.html',
            f'admin/{name}.html'
        ]



from typing import List, Sequence, Tuple, Union
from django.contrib.admin.views.main import ChangeList
from django import forms


__all__ = (
    'FieldsDefinition',
    'InjectableChangeList',
    'append_injector', 'after_injector',
    'make_form', 'make_view', 'make_admin_parts', 'make_admin_methods',
)

FieldsDefinition = List[
    Union[
        Tuple[str, forms.Field, callable],
        Tuple[str, forms.Field],
    ]
]


class InjectableChangeList(ChangeList):
    """ ChangeList view class with injection methods.
    """
    list_editable_injection_fields: FieldsDefinition = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.list_editable += [
            key for key, *_ in self.list_editable_injection_fields
        ]

        self.list_display = self.inject_list_editable_fields(
            self.list_display, self.list_editable_injection_fields
        )

    def inject_list_editable_fields(
        self, source: List[str], fields: FieldsDefinition
    ):
        source = source[:]

        for key, _, injector in fields:
            source = injector(source, key)

        return source


def normalize_injection_fields(
    fields: FieldsDefinition, default_injector: callable
):
    """ Normalizes fields definition.
    """
    return [
        (
            definition[0],
            definition[1],
            definition[2] if len(definition) > 2 else default_injector
        )
        for definition in fields
    ]


def append_injector(sources: List[str], field: str):
    """ Simple injector that injects field at the end of the list.
    """
    return sources + [field]


def after_injector(after_field: str):
    """ Factory that returns an injector that injects field after the given field.
    """
    def injector(sources: List[str], field: str):
        index = sources.index(after_field) + 1

        return sources[:index] + [field] + sources[index:]

    return injector


def make_form(fields: FieldsDefinition, bases: Sequence[type] = []):
    """ Generates form class for the given fields.
    """
    bases = list(bases)

    if forms.ModelForm not in bases:
        bases.append(forms.ModelForm)

    return type('ChangelistForm', tuple(bases), {
        key: field
        for key, field, *_ in fields
    })


def make_view(
    fields: FieldsDefinition,
    default_injector: callable = append_injector,
    base_view_class: type = InjectableChangeList,
):
    """ Generates view class for the given fields.
    """
    return type('InjectableChangelistView', (base_view_class,), {
        'list_editable_injection_fields': normalize_injection_fields(
            fields, default_injector
        ),
    })


def make_admin_parts(
    fields: FieldsDefinition,
    default_injector: callable = append_injector,
    base_view_class: type = InjectableChangeList,
    form_bases: Sequence[type] = [],
):
    """
    Generates changelist form and view classes for the given fields.

    Args:
        fields (FieldsDefinition): List of fields definitions to inject.
        default_injector (callable, optional): Default injector function. Defaults to append_injector.
        base_view_class (type, optional): Base view class. Defaults to InjectableChangeList.
        form_bases (Sequence[type], optional): Base classes for resulting form. Defaults to [].

    Returns:
        Tuple: Form and view classes.
    """
    return (
        make_form(fields, form_bases),
        make_view(fields, default_injector, base_view_class)
    )


def make_admin_methods(
    fields: FieldsDefinition,
    default_injector: callable = append_injector,
    base_view_class: type = InjectableChangeList,
    form_bases: Sequence[type] = [],
):
    """
    Generates admin editable field injection methods for the given fields.

    Args:
        fields (FieldsDefinition): List of fields definitions to inject.
        default_injector (callable, optional): Default injector function. Defaults to append_injector.
        base_view_class (type, optional): Base view class. Defaults to InjectableChangeList.
        form_bases (Sequence[type], optional): Base classes for resulting form. Defaults to [].

    Returns:
        Tuple: Two methods: `get_changelist_form` and `get_changelist_view`.
    """
    form, view = make_admin_parts(
        fields, default_injector, base_view_class, form_bases
    )

    return (
        # get_changelist_form
        lambda *args, **kwargs: form,
        # get_changelist
        lambda *args, **kwargs: view,
    )



from django.contrib.admin import FieldListFilter
from django.utils.translation import ugettext_lazy as _
from django.core.exceptions import ImproperlyConfigured


__all__ = ('TextInputFilter',)


class TextInputFilter(FieldListFilter):
    """Renders simple eq admin filter form with text input and submit button.
    """
    parameter_name = None
    template = 'admin/text_input_filter.html'
    field = None

    def __init__(self, field, request, params, model, model_admin, field_path):
        super().__init__(
            field, request, params, model, model_admin, field_path
        )

        self.field = field
        if self.parameter_name is None:
            self.parameter_name = self.field.name

        if self.parameter_name in params:
            value = params.pop(self.parameter_name)
            self.used_parameters[self.parameter_name] = value

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(**{self.parameter_name: self.value()})

    def value(self):
        """Returns the value (in string format) provided in the request's
        query string for this filter, if any. If the value wasn't provided then
        returns None.
        """
        return self.used_parameters.get(self.parameter_name, None)

    def has_output(self):
        return True

    def expected_parameters(self):
        """Returns the list of parameter names that are expected from the
        request's query string and that will be used by this filter.
        """
        return [self.parameter_name]

    def choices(self, cl):
        all_choice = {
            'selected': self.value() is None,
            'query_string': cl.get_query_string({}, [self.parameter_name]),
            'display': _('Clear'),
        }

        return ({
            'get_query': cl.params,
            'current_value': self.value(),
            'all_choice': all_choice,
            'parameter_name': self.parameter_name
        },)



from typing import Sequence
from functools import reduce

from django.db import models
from django.http.request import HttpRequest


__all__ = ('ListAnnotatorAdminMixin',)


class ListAnnotatorAdminMixin:
    """Mixin to easily add annotated fields to a select.

    Property `list_annotations` may be filled with an annotations definitions:

    Example:
        ```python
        class SomeAdmin:
            list_annotations = (
                ('participants', models.Count, 'participants_count'),
            )
        ```

        Where:

        - The field name to manupulate(might be None if function does not require it to be).
        - The second parameter is the callable that receives field and
            does something with it.
            Must return data that can be used as annotation.
        - Third parameter - is the resulting field name. May not be present,
            in that case the name will be `self.LIST_ANNOTATIONS_PREFIX + field`.
    """
    LIST_ANNOTATIONS_PREFIX: str = 'annotated_'
    list_annotations: Sequence = ()

    def _resolve_annotation_definition(self, map: dict, definition: Sequence):
        field, func, *_ = definition
        name = (
            definition[2]
            if len(definition) >= 3 else
            self.LIST_ANNOTATIONS_PREFIX + field
        )
        map[name] = func(field)

        return map

    def get_queryset(self, request: HttpRequest) -> models.QuerySet:
        q = super().get_queryset(request)
        annotations: dict = reduce(
            self._resolve_annotation_definition,
            self.get_list_annotations(request, q),
            {}
        )

        return q.annotate(**annotations)

    def get_list_annotations(
        self, request: HttpRequest, queryset: models.QuerySet
    ) -> Sequence:
        return self.list_annotations



from django.contrib import admin
from django.utils.translation import ugettext_lazy as _
from django.contrib.postgres.fields import ArrayField
from admin_totals.admin import ModelAdminTotals
from admin_numeric_filter.admin import NumericFilterModelAdmin
from django_better_admin_arrayfield.admin.mixins import DynamicArrayMixin
from django_better_admin_arrayfield.forms.fields import DynamicArrayField

from pxd_admin_extensions import (
    MultiDBModelAdmin, ListAnnotatorAdminMixin, FormTypeAdminMixin
)


__all__ = 'AllAdmin',


class AllAdmin(
    DynamicArrayMixin,
    ListAnnotatorAdminMixin,
    ModelAdminTotals,
    NumericFilterModelAdmin,
    FormTypeAdminMixin,
    MultiDBModelAdmin,
    admin.ModelAdmin
):
    """All the admin features in one place.
    """
    empty_value_display = '-----'

    formfield_overrides = {
        ArrayField: {'form_class': DynamicArrayField},
    }



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class RangeFilterExtensionConfig(AppConfig):
    name = 'pxd_admin_extensions.contrib.rangefilter'
    label = 'pxd_admin_extensions_rangefilter'
    verbose_name = _('Range filter admin extension')



from typing import Optional
from functools import wraps

from django.contrib import admin


__all__ = (
    'using_kwarg_factory',
    'UsableAdminBaseMixin',
    'UsableForeignMixin',
    'UsableAdminMixin',
    'MultiDBModelAdmin',
)


def using_kwarg_factory(func):
    """Simple method wrapper to get a default using parameter.

    Args:
        func (Callable): Wrapped method.
    """
    def method(self, *args, using=None, **kwargs):
        return func(
            self, *args, using=self.using if using is None else using, **kwargs
        )

    return wraps(func)(method)


class UsableAdminBaseMixin:
    """Simpe mixin to apply default database key to the queryset resolver.

    ModelAdmin attribute `using` is for controlling which database
    the QuerySet will be evaluated against if you are using more than
    one database. The value is the alias of a database, as defined in DATABASES.
    """
    using: Optional[str] = None

    def get_queryset(self, request):
        return super().get_queryset(request).using(self.using)


class UsableForeignMixin:
    """Mixin to add self `using` kwarg usage to ModelAdmin for foreign fields.
    """
    @using_kwarg_factory
    def formfield_for_foreignkey(self, *a, **k):
        return super().formfield_for_foreignkey(*a, **k)

    @using_kwarg_factory
    def formfield_for_manytomany(self, *a, **k):
        return super().formfield_for_manytomany(*a, **k)


class UsableAdminMixin(UsableAdminBaseMixin, UsableForeignMixin):
    pass


class MultiDBModelAdmin(UsableAdminMixin, admin.ModelAdmin):
    """Admin to search/create/change objects from non-default database.

    ModelAdmin attribute `using` is for controlling which database
    the QuerySet will be evaluated against if you are using more than
    one database. The value is the alias of a database, as defined in DATABASES.
    """
    def save_model(self, request, obj, form, change):
        obj.save(using=self.using)

    def delete_model(self, request, obj):
        obj.delete(using=self.using)



import json
from django import template
from django.core.serializers.json import DjangoJSONEncoder

from ..const import STATS_POSTFIX

register = template.Library()


@register.filter
def stats_json(value):
    """Filter to transform stats into filter-understandable json.
    """
    return json.dumps(value, cls=DjangoJSONEncoder)


@register.filter
def add_to_list(array, to_add):
    """List push filter implementation.
    """
    if not isinstance(array, (list, tuple)):
        array = [array]

    return array + [to_add]


@register.filter
def is_in_column(value):
    """Checks whether the current 'i' is in the right column(out of some amount).
    """
    columns, col, i = value

    return i % columns == col


@register.simple_tag(takes_context=True)
def get_stats_link(context, back=False):
    """Generates link to a stats admin or vice versa.
    """
    stats_postfix = STATS_POSTFIX
    request = context['request']
    path = request.path[:-request.path.endswith('/')]
    url = request.get_full_path()
    changed_path = path[:-len(stats_postfix)] if back else path + stats_postfix

    return url.replace(path, changed_path)



__all__ = 'FormTypeAdminMixin',


class FormTypeAdminMixin:
    """Mixin that let you use different forms for different object
    manupulation scenarios:

    - `changelist_form` - Special form for changelist admin view.
    - `add_form` - Special form for object creation.
    - `change_form` - Special form for object update mechanics.

    All those forms will fallback to a default one if none provided.
    """
    def get_form(self, request, obj=None, **kwargs):
        form = getattr(
            self, 'add_form' if obj is None else 'change_form', self.form
        ) or self.form

        if form is not None:
            kwargs.setdefault('form', form)

        return super().get_form(request, obj, **kwargs)

    def get_changelist_form(self, request, **kwargs):
        form = getattr(self, 'changelist_form', self.form) or self.form

        if form is not None:
            kwargs.setdefault('form', form)

        return super().get_changelist_form(request, **kwargs)



from datetime import datetime

from psycopg2.extras import DateTimeRange
from django.utils.translation import ugettext_lazy as _
from rangefilter.filter import DateTimeRangeFilter


__all__ = ('DateTimeRangeForRangeFilter',)


class DateTimeRangeForRangeFilter(DateTimeRangeFilter):
    """Date range filter for no-timezone dates.
    """
    def _make_query_filter(self, request, validated_data):
        gte: datetime = validated_data.get(self.lookup_kwarg_gte, None)
        lte: datetime = validated_data.get(self.lookup_kwarg_lte, None)

        return {
            f'{self.field_path}__overlap': DateTimeRange(
                gte.replace(tzinfo=None),
                lte.replace(tzinfo=None),
                bounds='[]'
            )
        }



__all__ = 'stat_legend', 'stat',


def stat_legend(key, title=None, **kwargs):
    """Simplifier for statistics legend definition.

    Args:
        key (str): Legend key.
        title (str, optional): Legend's title. Defaults to None.
    """
    return {'key': key, 'title': title or key, **kwargs}


def stat(
    title=None, legend=(), key=None, y=None, x=None, template_name=None,
    config={}
):
    """Statistics resolver method wrapper.
    Registers a new statistics block to display and with what inside.

    Args:
        title (str, optional): Block title. Defaults to None.
        legend (List[str], optional): List of field descriptors.
            Defaults to ().
        key (str, optional): Stat block unique key. Defaults to None.
        y (str, optional): Y axis key. Defaults to None.
        x (str, optional): X axis key. Defaults to None.
        template_name (str, optional): Django template name to use for block
            rendering.
            Defaults to None.
        config (dict, optional): Additional data to configure block.
            Defaults to {}.

    Example:
        ```python
        @register_stats(ExistingModelAdmin, model=ExistingModel)
        class ExistingModelStatsAdmin:
            @stat(_('Participants per domain'), (
                stat_legend('domain', _('Domain')),
                stat_legend('participants_count', _('Participants count')),
            ), config={'type': 'pie', 'labelType': 'percent'})
            def get_participants_per_domain(self, request, qs):
                return (
                    qs
                    .values('domain')
                    .annotate(participants_count=models.Count(
                        'participants__participant_id', distinct=True
                    ))
                    .order_by('domain')
                )
        ```

        By default every `stat` must return list of dicts with some valuable
        stats data. That will be then transformed using legend and config
        into a required chart.
    """
    assert len(legend) > 1, 'There must be at least 2 parameters to manage.'
    assert len(legend) < 4, 'We could match not more than 3 values at a time.'

    x = x or legend[0]['key']
    y = y or legend[1]['key']

    def wrapper(func):
        func.title = title
        func.template_name = template_name
        func.config = {'legend': legend, 'y': y, 'x': x, **config}
        func.key = key
        func.is_stat = True

        return func
    return wrapper



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class StatsConfig(AppConfig):
    name = 'pxd_admin_extensions.contrib.stats'
    label = 'pxd_admin_extensions_stats'
    verbose_name = _('Stats admin extension')



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class JetExtensionConfig(AppConfig):
    name = 'pxd_admin_extensions.contrib.jet'
    label = 'pxd_admin_extensions_jet'
    verbose_name = _('Jet admin extension')



__version__ = '0.1.2'

from .annotations import *
from .multidb import *
from .form_types import *
from .changelist_template_names import *

from .filters import *

VERSION = tuple(__version__.split('.'))



from django.contrib import admin

from .admin import StatsAdmin
from .const import STATS_POSTFIX


__all__ = (
    'create_simple_proxy_model',
    'register_stats',
)


ADMIN_STAT_ALTERNATIVES = {
    None: 'admin/stats_change_list.html',
    'admin/change_list.html': 'admin/stats_change_list.html',
    'admin_totals/change_list_totals.html': 'admin_totals/stats_change_list.html',
}


def create_simple_proxy_model(
    Model,
    postfix: str='Proxy',
    verbose_name: str=None,
    verbose_name_plural: str=None,
):
    """Factory to create "empty" proxy model.

    Args:
        Model (type): Base model to inherit from.
        postfix (str, optional): Postfix for model to have unique name.
            Defaults to 'Proxy'.
        verbose_name (str, optional): Model's verbose name. Defaults to None.
        verbose_name_plural (str, optional): Model's plural verbose name.
            Defaults to None.
    """
    v = verbose_name or Model._meta.verbose_name_raw
    vp = verbose_name_plural or Model._meta.verbose_name_plural

    class Meta:
        proxy = True
        verbose_name = v
        verbose_name_plural = vp

    return type(Model.__name__ + (postfix or 'Proxy'), (Model, ), {
        '__module__': Model.__module__,
        'Meta': Meta
    })


def register_stats(
    underlying_admin, registerer=admin.register, stats_admin=StatsAdmin,
    model=None, **kwargs
):
    """Factory to generate admin panel for statistics display.

    Args:
        underlying_admin (ModelAdmin): Current admin class.
        registerer (Callable, optional): Admin registerer.
            Defaults to admin.register.
        stats_admin (type, optional): Base class with admin statistics
            imlementation.
            Defaults to StatsAdmin.
        model (Model, optional): Model to add stats for. Defaults to None.
    """
    def wrapper(Class):
        postfix = STATS_POSTFIX
        Model = model or underlying_admin.model
        ProxyModel = create_simple_proxy_model(
            Model, 
            postfix=postfix,
            verbose_name=kwargs.pop('verbose_name', None),
            verbose_name_plural=kwargs.pop('verbose_name_plural', None)
        )
        Admin = type(
            underlying_admin.__name__,
            (stats_admin, Class, underlying_admin,),
            {
                'underlying_admin': underlying_admin,
                'underlying_admin': Model,
            }
        )
        underlying_admin.change_list_template = ADMIN_STAT_ALTERNATIVES.get(
            underlying_admin.change_list_template,
            underlying_admin.change_list_template
        )

        return registerer(ProxyModel, **kwargs)(Admin)

    return wrapper



from django.utils.translation import ugettext_lazy as _
from django.contrib.admin import SimpleListFilter


__all__ = ('MethodFilter',)


class MethodFilter(SimpleListFilter):
    """Simple admin filter to call queryset's method if selected as `yes`.
    """
    def lookups(self, request, model_admin):
        return (
            (None, _('No')),
            ('1', _('Yes')),
        )

    def queryset(self, request, queryset):
        if self.used_parameters.get(self.parameter_name) == '1':
            return getattr(queryset, self.method_name)()

        return queryset

    @classmethod
    def as_filter(cls, parameter_name: str, method: str = None, title=None):
        """Factory for easier filter generation.

        Args:
            parameter_name (str): Parameter to bind this filter.
            method (str, optional): Method to call. Defaults to None.
            title ([type], optional): Filters title. Defaults to None.
        """
        return type(parameter_name + 'Filter' + cls.__name__, (cls,), {
            'parameter_name': parameter_name,
            'method_name': method or parameter_name,
            'title': title or parameter_name
        })



from jet.admin import CompactInline

from ...multidb import UsableAdminMixin


__all__ = ('MultiDBCompactInline',)


class MultiDBCompactInline(UsableAdminMixin, CompactInline):
    """Compact inline for multidb usage.
    """
    pass



import os
import re
import setuptools


with open('README.md', 'r') as rf:
    with open('CHANGELOG.md', 'r') as cf:
        long_description = rf.read() + cf.read()


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()

    return re.search('__version__ = [\'"]([^\'"]+)[\'"]', init_py).group(1)


version = get_version('pxd_admin_extensions')


setuptools.setup(
    name='preusx-django-admin-extensions',
    version=version,
    author='Alex Tkachenko',
    author_email='preusx@gmail.com',
    license='MIT License',
    description='Utilities and extensions for django administration panel.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=(),
    extras_require={
        'contrib': (
            'django-admin-rangefilter==0.6.3',
            'django-admin-numeric-filter==0.1.6',
            'django-admin-easy==0.6',
            'django-better-admin-arrayfield==1.4.2',
        ),
        'docs': (
            'mkdocs==1.1.2',
            'mkdocstrings==0.14.0',
            'mkdocs-material==6.2.4',
            'mkdocs-include-markdown-plugin==3.3.0',
        ),
        'dev': (
            'pytest>=6.0,<7.0',
            'pytest-watch>=4.2,<5.0',
            'pytest-django>=4.3,<5.0',
            'django-environ==0.4.5',
            'django-stubs',
            'django>=2.2,<4',
            'twine',
        ),
    },
    include_package_data=True,
    packages=setuptools.find_packages(exclude=(
        'tests', 'tests.*',
        'example', 'example.*',
        'docs', 'docs.*',
    )),
    python_requires='>=3.6',
    classifiers=(
        'Development Status :: 2 - Pre-Alpha',

        'Programming Language :: Python :: 3',

        'Intended Audience :: Developers',
        'Topic :: Utilities',

        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ),
)



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class AllExtensionConfig(AppConfig):
    name = 'pxd_admin_extensions.contrib.all'
    label = 'pxd_admin_extensions_all'
    verbose_name = _('All contrib together app')



from django.contrib import admin
from pxd_admin_extensions import ChangeListCustomTemplateNameAdminMixin


__all__ = (
    'ViewOnlyAdminMixin',
    'StatsAdmin',
)


class ViewOnlyAdminMixin:
    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class StatsAdmin(
    ViewOnlyAdminMixin,
    ChangeListCustomTemplateNameAdminMixin,
    admin.ModelAdmin
):
    """Admin basement for stats interface.
    """
    change_list_template_name = 'stats_list'
    stat_display_default_template_name = 'admin/includes/stats/display.html'
    stat_columns_amount = 2
    underlying_admin = None
    underlying_model = None

    class Media:
        css = {'all': ('admin/stats/nv.d3.min.css',)}
        js = (
            'admin/stats/d3.min.js', 'admin/stats/nv.d3.min.js',
            'admin/stats/chart.js',
        )

    def get_stats_date_period(self, request, date_hierarchy=None):
        """Simple method to resolve date truncation basedon the date
        hierarchy filter.
        """
        date_hierarchy = date_hierarchy or self.date_hierarchy

        if date_hierarchy + '__day' in request.GET:
            return 'minute'
        if date_hierarchy + '__month' in request.GET:
            return 'day'
        if date_hierarchy + '__year' in request.GET:
            return 'month'

        return 'year'

    def collect_stats(self, request, qs):
        """Collects all the defined stat callers and executes them.
        """
        attributes = ((key, getattr(self, key)) for key in dir(self))
        stat_callables = (
            (key, attr) for key, attr in attributes
            if callable(attr) and hasattr(attr, 'is_stat') and attr.is_stat
        )
        key_normalized = (
            (
                attr.key if hasattr(attr, 'key') and attr.key else
                key[key.startswith('get_') and 4 or 0:],
                attr
            )
            for key, attr in stat_callables
        )

        return [
            {
                'key': k,
                'result': list(attr(request, qs)),
                'config': getattr(attr, 'config', None) or {},
                'template_name': (
                    getattr(attr, 'template_name', None) or
                    self.stat_display_default_template_name
                ),
                'title': getattr(attr, 'title', None) or k,
            }
            for k, attr in key_normalized
        ]

    def changelist_view(self, request, extra_context=None):
        response = super().changelist_view(
            request, extra_context=extra_context,
        )
        try:
            qs = response.context_data['cl'].queryset
        except (AttributeError, KeyError):
            return response

        response.context_data['stats'] = self.collect_stats(request, qs)
        response.context_data['stats_columns'] = list(range(min(
            self.stat_columns_amount, len(response.context_data['stats'])
        )))

        return response
